
class Renderer():
    pass

# These imports we want in this name space, but they use the stuff above!
from .plotly import Plotly, CrossPlot # this is not great namespacing TODO
from .tree_table import TreeTable