cangrejo  = {
    "CALI" : {
        "color" : "black",
        "range": [0, 16],
        "range_unit": "inch",
    },
    "CGR" : {
        "color" : "green",
        "range": [0, 150.00],
        "range_unit": "gAPI",
    },
    "LLS" : {
        "color" : "gray",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit": "ohm * meter",
    },
    "ILM" : {
        "color" : "blue",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "ILD" : {
        "color" : "black",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "LLD" : {
        "color" : "red",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "LLD" : {
        "color" : "red",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit":  "ohm * meter",
    },
    "LLD" : {
        "color" : "red",
        "range": [0.2, 2000],
        "scale": "log",
        "range_unit": "ohm * meter",
    },
    "NPHI" : {
        "color" : "#008080",
        "range": [0.45, -.15],
        "range_unit": "fraction",
    },
    "RHOB" : {
        "color" : "red",
        "range": [1.95, 2.95],
        "range_unit": "gram/centimeter**3",
    },
    "DT" : {
        "color" : "red",
        "range": [30, 200],
        "range_unit": "microsecond/foot",
    },
}
